import typing
import inspect


class LoadStrategy(object):
    """
    A description of how to load a dataframe.
    """
    def __init__(self, format: str=None, read_args: dict=None, use_dask: bool=False, read_method: typing.Callable=None,
                 uri: str=None, open_args: dict=None, opener: typing.Callable=None,
                 predefined=None, sql: str=None):
        """
        :param format:      File format, as a file extension, i.e. "csv", etc.
        :param read_args:   Arguments to send to the detected read_method.
        :param use_dask:    True to use dask and not pandas.
        :param read_method: Detected method, like pandas.read_csv(), etc.
        :param uri:         URL to open with opener()
        :param open_args:   Arguments to send to the detected opener() method.
        :param opener:      Method that opens a stream to read this data.
        :param predefined:  Predefined DataFrame to use instead of anything else.
        :param sql:         SQL, if data is coming from a database.
        """
        self.format = format
        self.read_args = read_args
        self.use_dask = use_dask
        self.read_method = read_method
        self.uri = uri
        self.open_args = open_args
        self.opener = opener
        self.predefined = predefined
        self.sql = sql

    @staticmethod
    def method_name(method: typing.Callable):
        """
        Work out package name and method name.

        :returns:  A tuple with
                0) name of module to import, or None
                1) qualified name of method to use
        """
        if not method:
            return None, None
        method_name = method.__name__
        module = method.__module__ if hasattr(method, "__module__") else None
        if module in ("pandas", "pandas.io.parsers.readers"):
            return "pandas", "pandas." + method_name
        if module == "dslibrary" or module.startswith("dslibrary."):
            return "dslibrary", "dslibrary." + method_name
        if module == "io" and method_name == "open":
            return None, "open"
        if module.startswith("dask.dataframe"):
            return "dask.dataframe", "dask.dataframe." + method_name
        return module, module + "." + method_name

    def _reader_supports_filename(self):
        """
        Some read functions (i.e. pandas.read_csv()) accept a filename.
        """
        reader_module = self.read_method.__module__ if hasattr(self.read_method, "__module__") else None
        if reader_module == "pandas" or reader_module.startswith("pandas"):
            return True
        if reader_module.startswith("dask."):
            return True
        return False

    def _reader_requires_filename(self):
        """
        Some read functions (i.e. dask.dataframe.read_csv()) require a filename and do not work with streams.
        """
        reader_module = self.read_method.__module__ if hasattr(self.read_method, "__module__") else None
        if reader_module.startswith("dask."):
            return True
        return False

    def _opener_can_be_simplified_to_filename(self):
        """
        Some combinations of 'opener' and 'open_args' translate into "just open this file".
        """
        if not self.opener:
            return True
        method_name = self.opener.__name__
        module = self.opener.__module__ if hasattr(self.opener, "__module__") else None
        if (module == "dslibrary" or module.startswith("dslibrary.")) and method_name == "open_resource":
            unrecognized_options = set(self.open_args or []) - {"storage_options"}
            if not unrecognized_options:
                return True
        if module == "io" and method_name == "open":
            return True

    def _can_skip_opener(self):
        """
        Code like 'pandas.read_csv(open('file.csv')) can be simplified, because pandas.read_csv() accepts a filename.
        This simplification is important when using bodo's JIT compiler.
        """
        if self._reader_supports_filename() and self._opener_can_be_simplified_to_filename():
            return True

    @staticmethod
    def repr_kwargs(kwargs: dict, comma_before: bool=True):
        out = ", ".join(f"{k}={repr(v)}" for k, v in (kwargs or {}).items())
        if comma_before and out:
            out = ", " + out
        return out

    def to_pandas_code(self, result_var: str="data") -> str:
        """
        Convert the strategy into Python code that uses pandas to load the dataframe.
        """
        code = ""
        if self.predefined:
            code += "import pandas\n"
            # TODO sort out arguments to to_json()
            code += f"{result_var} = pandas.DataFrame({repr(self.predefined.to_json())})\n"
        elif self.sql:
            mod_o, method_o = self.method_name(self.opener)
            if mod_o:
                code += f"import {mod_o}\n"
            connect_code = f"{method_o}({self.repr_kwargs(self.open_args, comma_before=False)})"
            # TODO: here, 'read_method' is assumed to be a self-contained local utility method - some kind of verification should happen
            code += inspect.getsource(self.read_method) + "\n"
            code += f"{result_var} = {self.read_method.__name__}({connect_code})\n"
        else:
            read_args = self.read_args
            mod_o, method_o = self.method_name(self.opener)
            if self._can_skip_opener():
                # don't use opener
                mod_o, method_o = None, None
                # merge open args to read args (i.e. storage_options for s3)
                read_args = dict(read_args or {})
                read_args.update(self.open_args or {})
            mod_r, method_r = self.method_name(self.read_method)
            if mod_o:
                code += f"import {mod_o}\n"
            if mod_r:
                code += f"import {mod_r}\n"
            if method_o:
                if self._reader_requires_filename():
                    raise ValueError(f"The method {method_r}() requires a filename, and it was not possible to simplify the supplied arguments to a filename")
                code_open = f"{method_o}({repr(self.uri)}{self.repr_kwargs(self.open_args)})"
            else:
                code_open = repr(self.uri)
            code += f"{result_var} = {method_r}({code_open}{self.repr_kwargs(read_args)})\n"
        return code


class CloseAfterIteration(object):
    """
    Wrapper for a pandas DataFrame iterator (i.e. read_csv(chunksize=...)), which makes sure an associated file stream
    gets closed.
    """
    def __init__(self, stream, iterator):
        self.stream = stream
        self.iterator = iterator
        self.closed = False

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return next(self.iterator)
        except StopIteration:
            self.close()
            raise

    def close(self):
        if not self.closed:
            self.stream.close()
            self.closed = True

    def tell(self):
        if not hasattr(self.stream, "tell"):
            return
        return self.stream.tell()

    def get_chunk(self, size=None):
        try:
            if hasattr(self.iterator, "get_chunk"):
                return self.iterator.get_chunk(size)
        except StopIteration:
            self.close()
            raise

    def __getitem__(self, item):
        """
        Support sub-indexing.
        """
        if hasattr(self.iterator, "__getitem__"):
            # supplied iterator appears to support sub-indexing
            return CloseAfterIteration(self.stream, self.iterator[item])
        # wrap iterator with sub-indexing.
        return ApplySubindexing(self, item)


class ApplySubindexing(object):
    """
    Apply a slice to an iterator.
    """
    def __init__(self, parent, subindex: slice):
        if not isinstance(subindex, slice):
            raise ValueError("Only 'slice' is supported here.")
        self.parent = parent
        self.subindex = subindex
        self._wait = subindex.start or 0
        self._pos = 0

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            v = next(self.parent)
            if self.subindex.stop and self._pos >= self.subindex.stop:
                self.close()
                raise StopIteration
            self._pos += 1
            if self._wait <= 0:
                break
            self._wait -= 1
        self._wait = self.subindex.step - 1 if self.subindex.step else 0
        return v

    def close(self):
        if hasattr(self.parent, "close"):
            self.parent.close()

    def tell(self):
        if hasattr(self.parent, "tell"):
            return self.parent.tell()


def load_dataframe_from_sql(db_conn, sql: str, use_dask: int=0):
    """
    Given a database connection, load a dataframe.

    NOTE: this method must be self-contained (i.e. include all its own imports) so that it can be sent elsewhere
    for execution.
    """
    try:
        import pandas
        c = db_conn.cursor()
        c.execute(sql)
        col_names = [col[0] for col in c.description]
        df = pandas.DataFrame(data=c, columns=col_names)
        # TODO here we are, dutifully moving a too-big dataframe to dask, is it worth it?
        if use_dask:
            import dask.dataframe
            mem_size = df.memory_usage().sum()
            if mem_size > use_dask:
                partition_size = max(use_dask, 40000000)
                df = dask.dataframe.from_pandas(df, npartitions=min(mem_size // partition_size, 2))
        return df
    finally:
        db_conn.close()



import unittest
import pandas
import mock

from dslibrary.utils.connect import _process_params, dataframe_to_sql, db_conn_flavor, connect_psycopg2, \
    connect_to_database, enquote_sql_identifier


class TestConnect(unittest.TestCase):
    def test__process_params(self):
        r = _process_params("mysql://u:p@host:1234/database", extra=1)
        assert r == {
            'host': 'host',
            'port': 1234,
            'database': 'database',
            'username': 'u',
            'password': 'p',
            'extra': 1
        }, r
        r = _process_params("xyz://u@host")
        assert r == {
            'host': 'host',
            'port': None,
            'database': '',
            'username': 'u',
            'password': ''
        }, r
        r = _process_params("sqlite:path/to/file")
        assert r == {
            'host': '',
            'port': None,
            'database': 'path/to/file',
            'username': '',
            'password': ''
        }, r
        r = _process_params("x://host/db", database="db2")
        assert r == {
            'host': 'host',
            'port': None,
            'database': 'db2',
            'username': '',
            'password': ''
        }, r

    def test_dataframe_to_sql(self):
        df = pandas.DataFrame({"x": [1, 2, 3], "y": [1.0, 2.0, 4.0], "z": ["one", "two", "three"]})
        sqls = list(dataframe_to_sql(df, "t1"))
        self.assertEqual(sqls, [
            ('DROP TABLE IF EXISTS t1', []),
            ('CREATE TABLE IF NOT EXISTS t1 (x INTEGER, y DOUBLE PRECISION, z VARCHAR(8))', []),
            ('INSERT INTO t1 (x, y, z) VALUES (%s, %s, %s), (%s, %s, %s), (%s, %s, %s)', [1, 1.0, 'one', 2, 2.0, 'two', 3, 4.0, 'three'])
        ])

    def test_dataframe_to_sql__bigquery(self):
        df = pandas.DataFrame({"x": [1, 2, 3], "y": [1.0, 2.0, 4.0], "z": ["one", "two", "three"]})
        sqls = list(dataframe_to_sql(df, "t1", flavor="bigquery"))
        self.assertEqual(sqls, [
            ('DROP TABLE IF EXISTS t1', []),
            ('CREATE TABLE IF NOT EXISTS t1 (x INT64, y NUMERIC, z STRING(8))', []),
            ('INSERT INTO t1 (x, y, z) VALUES (%s, %s, %s), (%s, %s, %s), (%s, %s, %s)', [1, 1.0, 'one', 2, 2.0, 'two', 3, 4.0, 'three'])
        ])

    def test_enquote_sql_identifier(self):
        # no flavor
        self.assertEqual('a', enquote_sql_identifier("a"))
        self.assertEqual('"a 1"', enquote_sql_identifier("a 1"))
        self.assertEqual('"a!"', enquote_sql_identifier("a!"))
        self.assertEqual('a.b', enquote_sql_identifier("a.b"))
        self.assertEqual('a."b!"', enquote_sql_identifier("a.b!"))
        # postgres
        self.assertEqual('a', enquote_sql_identifier("a", flavor="postgres"))
        self.assertEqual('"A"', enquote_sql_identifier("A", flavor="postgres"))
        # mysql
        self.assertEqual('a', enquote_sql_identifier("a", flavor="mysql"))
        self.assertEqual('A', enquote_sql_identifier("A", flavor="mysql"))
        self.assertEqual('`A!`', enquote_sql_identifier("A!", flavor="mysql"))

    def test_db_conn_flavor(self):
        # artificially set property
        class A(object):
            _flavor = "xyz"
        self.assertEqual(db_conn_flavor(A()), "xyz")
        # name of class
        class MyPostGresDriver(object):
            pass
        self.assertEqual(db_conn_flavor(MyPostGresDriver()), "postgres")

    def test_sniff__pg(self):
        strategy = connect_to_database("postgres://u:p@host:1234/database", sniff_only=True)
        self.assertEqual(strategy.open_args, {'dsn': 'dbname=database host=host password=p port=1234 user=u'})
        import psycopg2
        self.assertIs(strategy.opener, psycopg2.connect)

    def test_pg_2(self):
        import dslibrary
        class MyPg(object):
            def set_isolation_level(self, n):
                pass
        def conn(**kwargs):
            self.assertEqual(kwargs, {'dsn': 'dbname=postgres host=HOST password=xyz port=5432 user=postgres'})
            return MyPg()
        with mock.patch("psycopg2.connect", conn):
            c = dslibrary.get_sql_connection('postgres://HOST:5432/postgres', password="xyz", **{'username': 'postgres'})
            assert isinstance(c, MyPg)

    def test_connect_psycopg2_valid_dsn(self):
        class MyPg(object):
            def set_isolation_level(self, n):
                pass
        def conn(**kwargs):
            self.assertEqual(kwargs, {'dsn': 'dbname=db1 host=host password=p port=1234 user=u', 'service': 's1'})
            return MyPg()
        with mock.patch("psycopg2.connect", conn):
            connect_psycopg2(host="host", port=1234, username="u", password="p", database="db1", a=123, z="def",
                             service="s1")




from urllib.parse import urlparse, unquote, urlunparse, quote
import re
import typing


def connect_to_database(uri: str, library: str=None, for_write: bool=False, **kwargs):
    """
    Try, by various means, to connect to a database.

    :param uri:        URI of database.
    :param library:     A specific library to use, if available
    :param for_write:   Whether to enable write operations.
    :param sniff_only:  True to return a ConnectStrategy instead of a connection.
    :returns:  a DBI driver
    """
    # determine flavor of connection and work out which libraries to try
    libraries = []
    if library:
        libraries.append(library)
    uri_parts = uri.split(":")
    if uri_parts[0] == "jdbc":
        uri_parts = uri_parts[1:]
    if uri_parts[0] in ("postgres", "postgresql"):
        libraries.append("psycopg2")
    if uri_parts[0] == "mysql":
        libraries.append("pymysql")
    if uri_parts[0] == "mssql":
        libraries.append("pymssql")
    if uri_parts[0] == "sqlite":
        libraries.append("sqlite")
    params = _process_params(uri, **kwargs)
    params.pop("scheme", None)
    for lib in libraries:
        if lib in METHODS:
            conn = METHODS[lib](**params, for_write=for_write)
            if conn:
                return conn
    raise ValueError(f"No driver installed/supported for {uri_parts[0]}")


class ConnectStrategy(object):
    """
    How to connect to a database.
    """
    def __init__(self, opener: typing.Callable, open_args: dict=None):
        """
        :param opener:      The connection method.
        :param open_args:   Arguments to pass.
        """
        self.opener = opener
        self.open_args = open_args


def _process_params(uri: str, **kwargs):
    """
    Use a URI to provide most of the needed values for connecting to a database.  'kwargs' supplies overrides.
    :param uri:         A URI which can supply most of the needed values.
    :param kwargs:      Additional overrides.
    :return:        A {} with all the extracted, named connection parameters.
    """
    if uri.startswith("jdbc:"):
        uri = uri[5:]
    if ":" not in uri:
        host, path = uri, ""
    else:
        parsed = urlparse(uri)
        host = parsed.netloc
        path = parsed.path
        if path.startswith("/"):
            path = path[1:]
    user = pwd = ""
    if "@" in host:
        # NOTE: '@' can occur in 'user', i.e. for Google's git repositories
        u_p, host = re.split(r'@(?!.*@)', host)
        if ":" in u_p:
            user, pwd = u_p.split(":")
        else:
            user, pwd = u_p, ""
    if ":" in host:
        host, port = host.split(":", maxsplit=1)
    else:
        port = None
    params = {
        "host": host,
        "port": int(port) if port else None,
        "database": path,
        "username": user,
        "password": pwd
    }
    params.update(kwargs)
    return params


def connect_pymysql(host, port=None, username=None, password=None, database=None, autocommit=True, for_write: bool=False, sniff_only: bool=False, **kwargs):
    try:
        import pymysql
    except ImportError:
        return
    port = port or 3306
    params = dict(user=username, password=password, host=host, port=port, database=database or None, autocommit=autocommit, **kwargs)
    if sniff_only:
        return ConnectStrategy(opener=pymysql.connect, open_args=params)
    conn = pymysql.connect(**params)
    if not for_write:
        _set_conn_read_only(conn)
    conn._flavor = "mysql"
    return conn


def connect_pymssql(host, port=None, username=None, password=None, database=None, autocommit=True, for_write: bool=False, sniff_only: bool=False, **kwargs):
    try:
        import pymssql
    except ImportError:
        raise Exception("MS SQL library not installed")
    more = {}
    if "timeout" in kwargs:
        more["timeout"] = int(kwargs["timeout"])
    params = dict(
        server=host,
        user=username,
        password=password,
        database=database,
        autocommit=autocommit,
        **more
    )
    if sniff_only:
        return ConnectStrategy(opener=pymssql.connect, open_args=params)
    conn = pymssql.connect(**params)
    if not for_write:
        _set_conn_read_only(conn)
    conn._flavor = "mssql"
    return conn


def connect_psycopg2(host, port=None, username=None, password=None, database=None, autocommit=True, for_write: bool=False, sniff_only: bool=False, **kwargs):
    try:
        import psycopg2
        import psycopg2.extensions
    except ImportError:
        return
    # see: https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS
    allowed_dsn = {
        #"host", "hostaddr", "port", "dbname", "user", "password", "passfile",
        "channel_binding", "connect_timeout", "client_encoding",
        "options", "application_name", "fallback_application_name",
        "keepalives", "keepalives_idle", "keepalives_interval",
        "keepalives_count",
        "tcp_user_timeout", "tty", "replication", "gssencmode", "sslmode", "requiressl",
        "sslcompression", "sslcert", "sslkey", "sslpassword", "sslrootcert", "sslcrl", "sslcrldir",
        "sslcni", "requirepeer", "ssl_min_protocol_version", "ssl_max_protocol_version",
        "krbsrvname", "gsslib", "service", "target_session_attrs"
    }
    kwargs = {k: v for k, v in kwargs.items() if k in allowed_dsn}
    dsn_parts = {"host": host, "port": port or 5432, "dbname": database or "postgres"}
    if username:
        dsn_parts["user"] = username
    if password:
        dsn_parts["password"] = password
    pg_args_str = " ".join("%s=%s" % (k, v) for k, v in sorted(dsn_parts.items()))
    params = dict(dsn=pg_args_str, **kwargs)
    if sniff_only:
        return ConnectStrategy(opener=psycopg2.connect, open_args=params)
    conn = psycopg2.connect(**params)
    if autocommit:
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    if not for_write:
        _set_conn_read_only(conn)
    try:
        conn._flavor = "postgres"
    except:
        pass
    return conn


def connect_sqlite(host, database=None, for_write: bool=False, sniff_only: bool=False, **kwargs):
    try:
        import sqlite3
    except ImportError:
        return
    for ignore in ["port", "username", "password", "database", "autocommit"]:
        kwargs.pop(ignore, None)
    if for_write:
        params = dict(database=host, **kwargs)
    else:
        params = dict(database=f'file:{host or database}?mode=ro', uri=True, **kwargs)
    if sniff_only:
        return ConnectStrategy(opener=sqlite3.connect, open_args=params)
    conn = sqlite3.connect(**params)
    conn._flavor = "sqlite"
    return conn


def _set_conn_read_only(conn):
    """
    Try to enforce a read-only connection
    """
    try:
        conn.cursor().execute("set transaction read only")
    except:
        print(f"couldn't set read-only mode on connection: {conn.__class__.__name__}")


def enquote_sql_identifier(name, allow_separator: bool=True, flavor: str=None):
    name = str(name)
    # '.' in name will be treated as separator between schema and table name
    if "." in name and allow_separator:
        return ".".join(map(enquote_sql_identifier, name.split('.')))
    # postgres needs quoting of uppercase
    if flavor == "postgres":
        if re.search(r'[^a-z0-9_]', name) or name[:1].isdigit():
            return f'"{name}"'
        return name
    # all other engines
    if re.search(r'[^A-Za-z0-9_]', name) or name[:1].isdigit():
        if flavor == "mysql":
            return f'`{name}`'
        return f'"{name}"'
    return name


def dataframe_to_sql(dataframe, table_name: str, append: bool=False, flavor: str=None):
    """
    Generate SQL statements to create a table and add data.
    :param dataframe:   A dataframe to render into SQL.
    :param table_name:  Name of table to create or extend.
    :param append:      True to append, False to replace.
    :param flavor:      Name of a particular type of database, to adjust details of SQL format.
    :returns a generator of (sql, parameter list)
    """
    enquote = lambda name, **k: enquote_sql_identifier(name, flavor=flavor, **k)
    if not append:
        yield "DROP TABLE IF EXISTS %s" % enquote(table_name), []
    # TODO expand to cover more types
    if flavor == "bigquery":
        dtype_to_sql = {"int32": "INT64", "int64": "INT64", "float64": "NUMERIC", "datetime64": "TIMESTAMP"}
        str_type = "STRING"
    else:
        dtype_to_sql = {"int32": "INTEGER", "int64": "INTEGER", "float64": "DOUBLE PRECISION", "datetime64": "TIMESTAMP"}
        str_type = "VARCHAR"
    cols = list(dataframe.columns)
    types = []
    for col, dtype in zip(cols, dataframe.dtypes):
        sql_type = dtype_to_sql.get(str(dtype))
        if not sql_type:
            max_len = int(dataframe[col].str.len().max()) + 3
            sql_type = f"{str_type}({max_len:d})"
        types.append(sql_type)
    rows = dataframe.itertuples(index=False, name=None)
    fields = ", ".join("%s %s" % (enquote(col, allow_separator=False), t) for col, t in zip(cols, types))
    yield "CREATE TABLE IF NOT EXISTS %s (%s)" % (enquote(table_name), fields), []
    ins0 = "INSERT INTO %s (%s) VALUES " % (enquote(table_name), ", ".join(map(enquote, cols)))
    sqls = []
    params = []
    for row in (rows or []):
        sql = "(%s)" % ", ".join(map(lambda v: "%s", row))
        sqls.append(sql)
        params += list(row)
        if len(sqls) >= 1000:
            yield "%s%s" % (ins0, ", ".join(sqls)), params
            sqls.clear()
            params.clear()
    if sqls:
        yield "%s%s" % (ins0, ", ".join(sqls)), params


def db_conn_flavor(connector) -> str:
    """
    Determine connector flavor.
    """
    if hasattr(connector, "_flavor"):
        return connector._flavor
    name = connector.__class__.__name__.lower()
    if "postg" in name or "psycopg" in name:
        return "postgres"


METHODS = {
    "pymysql": connect_pymysql,
    "psycopg2": connect_psycopg2,
    "pymssql": connect_pymssql,
    "sqlite": connect_sqlite
}

# TODO BigQuery



import unittest
import tempfile
import shutil
import io
import mock

from dslibrary.utils.file_utils import FileOpener, connect_to_filesystem, write_stream_with_read_on_close, \
    adapt_fsspec_storage_options


class TestFileUtils(unittest.TestCase):

    def test_FileOpener_local_file(self):
        tmp_f = tempfile.mkdtemp()
        fo = FileOpener(tmp_f)
        with fo.open("x", mode="w") as f_w:
            f_w.write("abc")
        with fo.open("x", mode="r") as f_r:
            self.assertEqual(f_r.read(), "abc")
        shutil.rmtree(tmp_f)

    def test_FileOpener_s3(self):
        tmp_f = tempfile.mkdtemp()
        fo = FileOpener(tmp_f)
        def opener(uri, **kwargs):
            self.assertEqual(uri, "s3://bucket/file.csv")
            self.assertEqual(kwargs, {'mode': 'r', 'key': 'K', 'secret': 'S'})
            return io.StringIO("x")
        with mock.patch("fsspec.open", opener):
            with fo.open("s3://bucket/file.csv", mode="r", access_key="K", secret_key="S") as f_r:
                d = f_r.read()
        self.assertEqual(d, "x")
        shutil.rmtree(tmp_f)

    def test_write_stream_with_read_on_close(self):
        log = []
        f_w = write_stream_with_read_on_close('w', 'r', on_close=lambda fh: log.append(fh.read()))
        f_w.write("abc")
        f_w.write("def")
        f_w.close()
        assert log == ["abcdef"]

    def test_connect_to_filesystem(self):
        tmp_f = tempfile.mkdtemp()
        fs = connect_to_filesystem(tmp_f, for_write=True)
        with fs.open("x", mode="w") as f_w:
            f_w.write("abc")
        with fs.open("x", mode="r") as f_r:
            self.assertEqual(f_r.read(), "abc")
        self.assertEqual(fs.ls(), [{'name': 'x', 'size': 3, 'type': 'file'}])
        self.assertEqual(fs.stat("x"), {'name': 'x', 'size': 3, 'type': 'file'})
        assert fs.exists("x") is True
        assert fs.exists("y") is False
        shutil.rmtree(tmp_f)

    def test_adapt_fsspec_storage_options(self):
        self.assertEqual(adapt_fsspec_storage_options({"access_key": "K"}), {'storage_options': {'key': 'K'}})
        self.assertEqual(adapt_fsspec_storage_options({"key": "K"}), {'storage_options': {'key': 'K'}})
        self.assertEqual(adapt_fsspec_storage_options({"storage_options": {"x": 1}, "access_key": "K", "z": 2}), {'storage_options': {'key': 'K', 'x': 1}, 'z': 2})




def connect_to_csp(service: str, for_write: bool=True, **kwargs):
    """
    Connect to various CSP services.
    :param service:         Name of a service on some CSP.
    :param for_write:       Ignored for the time being.
    :param kwargs:          Credentials and other options.
    :return: A native CSP connector, i.e. a boto3 client.
    """
    if service in ("s3", "sagemaker", "lambda"):
        import boto3
        access = {}
        for i, o in {
            "key": "aws_access_key_id", "secret": "aws_secret_access_key",
            "access_key": "aws_access_key_id", "secret_key": "aws_secret_access_key",
            "aws_access_key_id": "aws_access_key_id", "aws_secret_access_key": "aws_secret_access_key",
            "region": "region_name", "region_name": "region_name"
        }.items():
            if i in kwargs:
                access[o] = kwargs[i]
        return boto3.client(service, **access)




"""
Get a quick overview of a file with basic information like column and row counts.
"""
import dslibrary
import copy


class DataPreview(object):
    """
    See data_preview().
    """
    def __init__(self, n_rows, n_cols, col_names, col_types, sample_rows):
        self.n_rows = n_rows
        self.n_cols = n_cols
        self.col_names = col_names
        self.col_types = col_types
        self.sample_rows = sample_rows

    def to_json(self):
        return self.__dict__


def data_preview(access: dict, preview_size=20000, capture_rows=10, max_cols=100):
    """
    Gather basic preview information about a data source.

    :param access:      Arguments for dslibrary.load_dataframe()
    :param preview_size: Maximum number of bytes to read from files.
    :param capture_rows: Number of rows to sample.
    :param max_cols:     Maximum number of columns to consider.
    :returns:  A DataPreview instance.
    """
    access = copy.deepcopy(access)
    if "sql_table" in access:
        raise NotImplementedError("not supported yet: sql_table")
    if "sql_source" in access:
        flavor = access.get("sql_flavor")
        # count rows
        access['resource_name'] = f"select count(*) as row_count from ({access['resource_name']})"
        df = dslibrary.load_dataframe(**access)
        n_rows = df.row_count[0]
        # get sample rows
        if flavor == "mssql":
            access['resource_name'] = f"select top {capture_rows} * from ({access['resource_name']})"
        else:
            access['resource_name'] = f"select * from ({access['resource_name']}) limit {capture_rows}"
        df = dslibrary.load_dataframe(**access)
    else:
        if "format_options" not in access:
            access["format_options"] = {}
            access["format_options"]["chunksize_bytes"] = preview_size
        df_stream = dslibrary.load_dataframe(**access)
        df = next(df_stream)
        try:
            next(df_stream)
            truncated = True
        except StopIteration:
            truncated = False
        n_rows = len(df)
        if truncated:
            tot_size = _size(access)
            if tot_size:
                n_rows = int(n_rows * tot_size / min(tot_size, preview_size))
            else:
                n_rows = -1
    captured = list(map(lambda row: row[:max_cols], df.head(capture_rows).itertuples(name=None, index=False)))
    col_names = list(df.columns)
    col_dtypes = list(map(str, df.dtypes))
    return DataPreview(n_rows, len(col_names), col_names[:max_cols], col_types=col_dtypes[:max_cols], sample_rows=captured)


def _size(access: dict):
    with dslibrary.open_resource(**access, mode='rb') as f:
        if f.seekable():
            f.seek(0, 2)
            return f.tell()
